#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`lsp`
==================

.. module:: lsp
    :platform: Unix, Windows
    :synopsis:

.. moduleauthor:: hbldh <henrik.blidh@nedomkull.com>

Created on 2015-11-13

"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

import six
import numpy as np
from scipy.signal import spectrogram

from lspopt.data import C, WEIGHTS, f_h


def lspopt(n, c_parameter=20.0):
    """Calculate the multitaper windows described in [1], given window length and C parameter.

    Parameters
    ----------
    n : int
        Length of multitaper windows
    c_parameter : float
        The parameter `c` in [1]. Default is 20.0

    Returns
    -------
    H : ndarray
        Multitaper windows, of size [n_windows x n].
    w : ndarray
        Array of taper window weights.

    References
    ----------
    ...[1] Hansson-Sandsten, M. (2011). Optimal multitaper Wigner spectrum
           estimation of a class of locally stationary processes using Hermite functions.
           EURASIP Journal on Advances in Signal Processing, 2011, 10.

    """
    k = int(round((c_parameter - 1) * 10))
    if c_parameter != C[k]:
        print("Using c={0} instead of desired {1}.".format(C[k], c_parameter))

    weights = WEIGHTS[:, k][np.nonzero(WEIGHTS[:, k])][:10]
    weights /= np.sum(weights)
    K = len(weights)

    t1 = np.arange(-(n / 2) + 1, (n / 2) + 0.1, step=1.0) / f_h(n, K)
    h = np.vstack((np.ones((n,)), 2 * t1))
    for i in six.moves.range(1, K - 1):
        h = np.vstack((h, (2 * t1 * h.T[:, i]) - 2 * i * h.T[:, i - 1]))

    H = h.T * np.outer(np.exp(-(t1 ** 2) / 2), np.ones((K,), "float"))
    H /= np.array([np.linalg.norm(x) for x in H.T])

    return H.T.copy(), weights


def spectrogram_lspopt(
    x,
    fs=1.0,
    c_parameter=20.0,
    nperseg=256,
    noverlap=None,
    nfft=None,
    detrend="constant",
    return_onesided=True,
    scaling="density",
    axis=-1,
):
    """Convenience method for calling :py:meth:`scipy.signal.spectrogram` with
    :py:mod:`lspopt` multitaper windows.

    Parameters
    ----------
    x : array_like
        Time series of measurement values
    c_parameter : float
        The parameter `c` in [1].

    For all other parameters, see the :py:meth:`scipy.signal.spectrogram`
    docstring.

    Returns
    -------
    f : ndarray
        Array of sample frequencies.
    t : ndarray
        Array of segment times.
    Sxx : ndarray
        Spectrogram of x. By default, the last axis of Sxx corresponds to the
        segment times.

    Notes
    -----
    An appropriate amount of overlap will depend on the choice of window
    and on your requirements. In contrast to welch's method, where the entire
    data stream is averaged over, one may wish to use a smaller overlap (or
    perhaps none at all) when computing a spectrogram, to maintain some
    statistical independence between individual segments.

    References
    ----------
    ...[1] Hansson-Sandsten, M. (2011). Optimal multitaper Wigner spectrum
           estimation of a class of locally stationary processes using Hermite functions.
           EURASIP Journal on Advances in Signal Processing, 2011, 10.

    Examples
    --------
    >>> from scipy.signal import chirp
    >>> from lspopt import lsp
    >>> import matplotlib.pyplot as plt

    Generate a test signal.

    >>> fs = 10e3
    >>> N = 1e5
    >>> amp = 2 * np.sqrt(2)
    >>> noise_power = 0.001 * fs / 2
    >>> time = np.arange(N) / fs
    >>> freq = np.linspace(1e3, 2e3, N)
    >>> x = (amp * chirp(time, 1e3, 2.0, 6e3, method='quadratic') +
    >>>      np.random.normal(scale=np.sqrt(noise_power), size=time.shape))

    Compute and plot the spectrogram.

    >>> f, t, Sxx = lsp.spectrogram_lspopt(x, fs=fs, c_parameter=20.0)
    >>> plt.pcolormesh(t, f, Sxx)
    >>> plt.ylabel('Frequency [Hz]')
    >>> plt.xlabel('Time [sec]')
    >>> plt.show()

    """
    # Less overlap than welch, so samples are more statisically independent
    if noverlap is None:
        noverlap = nperseg // 8

    H, taper_weights = lspopt(n=nperseg, c_parameter=c_parameter)
    S_out = None
    # Call spectrogram method for each taper window.
    for taper_window, taper_weight in zip(H, taper_weights):
        f, t, Pxx = spectrogram(
            x,
            fs=fs,
            window=taper_window,
            nperseg=nperseg,
            noverlap=noverlap,
            nfft=nfft,
            detrend=detrend,
            return_onesided=return_onesided,
            scaling=scaling,
            axis=axis,
        )
        if S_out is None:
            S_out = taper_weight * Pxx
        else:
            S_out += taper_weight * Pxx
    return f, t, S_out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import


from lspopt.lsp import lspopt, spectrogram_lspopt
from lspopt.__version__ import __version__, version

__all__ = ["lspopt", "spectrogram_lspopt"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

.. moduleauthor:: hbldh <henrik.blidh@nedomkull.com>

Created on 2015-11-13

"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

from pkg_resources import resource_filename

import numpy as np

__all__ = ["C", "WEIGHTS", "f_h"]

# An array of C parameter values for which weights have been pre-calculated.
C = np.load(resource_filename("lspopt.data", "c.npy")).flatten()
# The pre-calculated Hermite polynomial coefficients
# for the C parameter values above.
WEIGHTS = np.load(resource_filename("lspopt.data", "weights.npy"))


def f_h(n, k):
    """Returns f_h value.

    :param n: Window length of multitaper windows.
    :type n: int
    :param k: Length of non-zero Hermite polynomial coefficient array.
    :type k: int
    :return: The f_h value.
    :rtype: float

    """
    return n / _K_TO_VALUE_.get(k)


# Given length of Hermite polynomial coefficient array, return
# a value to divide N with.
_K_TO_VALUE_ = {
    1: 5.4,
    2: 6.0,
    3: 7.3,
    4: 8.1,
    5: 8.7,
    6: 9.3,
    7: 9.8,
    8: 10.3,
    9: 10.9,
    10: 11.2,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`utils`
==================

.. module:: utils
    :platform: Unix, Windows
    :synopsis:

.. moduleauthor:: hbldh <henrik.blidh@nedomkull.com>

Created on 2015-11-15, 22:47

"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

import numpy as np
from scipy.linalg import sqrtm


def create_lsp_realisation(N, c, Fs, random_seed=None):
    """Create a realisation of a locally stationary process (LSP).

    For details, see section 2 in [1].

    Parameters
    ----------
    N : int
        Length of the process to generate.
    c: float
        Measure of stationarity of the process to generate.
    Fs: float
        Frequency.
    random_seed: int
        Random seed to apply before generating process.

    Returns
    -------
    x : ndarray
        The process realisation.
    H : ndarray
        The process-generating matrix H.
    Rx : ndarray
        The covariance matrix R_x(t, s).

    References
    ----------
    ...[1] Hansson-Sandsten, M. (2011). Optimal multitaper Wigner spectrum
           estimation of a class of locally stationary processes using Hermite functions.
           EURASIP Journal on Advances in Signal Processing, 2011, 10.

    """
    r_x = lambda a, b: _q((a + b) / 2) * _r(a - b, c)
    return _create_realisation(r_x, N, Fs, random_seed)


def create_lscp_realisation(N, c, Fs, m, d, random_seed=None):
    """Create a realisation of a locally stationary chirp process.

    For details, see section 2 in [1].

    Parameters
    ----------
    N : int
        Length of the process to generate
    c: float
        Measure of stationarity of the process to generate.
    m: float
        Chirp frequency.
    d: float
        Start of the chirp frequency.
    random_seed: int
        Random seed to apply before generating process.

    Returns
    -------
    x : ndarray
        The process realisation.
    H : ndarray
        The process-generating matrix H.
    Rx : ndarray
        The covariance matrix R_x(t, s).

    References
    ----------
    ...[1] Hansson-Sandsten, M. (2011). Optimal multitaper Wigner spectrum
           estimation of a class of locally stationary processes using Hermite functions.
           EURASIP Journal on Advances in Signal Processing, 2011, 10.

    """
    r_x = (
        lambda a, b: _q((a + b) / 2)
        * _r(a - b, c)
        * np.exp(1j * m * (a - b) * ((a + b) / 2 - d))
    )
    return _create_realisation(r_x, N, Fs, random_seed)


def create_mlsp_realisation(N, c_vector, Fs, random_seed=None):
    """Create a realisation of a Multicomponent locally stationary process.

    For details, see section 2 in [1].

    Parameters
    ----------
    N : int
        Length of the process to generate
    c_vector: list, tuple
        A list of stationarity measure values of the process to generate.
    Fs: float
        Frequency.
    random_seed: int
        Random seed to apply before generating process.

    Returns
    -------
    x : ndarray
        The process realisation.
    H : ndarray
        The process-generating matrix H.
    Rx : ndarray
        The covariance matrix R_x(t, s).

    References
    ----------
    ...[1] Hansson-Sandsten, M. (2011). Optimal multitaper Wigner spectrum
           estimation of a class of locally stationary processes using Hermite functions.
           EURASIP Journal on Advances in Signal Processing, 2011, 10.

    """
    r_x = lambda a, b: np.sum([_q((a + b) / 2) * _r(a - b, c) for c in c_vector])
    return _create_realisation(r_x, N, Fs, random_seed)


def _create_realisation(r_x, N, Fs, random_seed):
    # Sampling vector.
    t_vector = np.arange(-(N / 2), N / 2, 1) / Fs

    # Create real or complex matrix depending on covariance function output.
    if np.iscomplex(r_x(t_vector[0], t_vector[-1])):
        Rx = np.zeros((N, N), "complex")
    else:
        Rx = np.zeros((N, N), "float")

    # Calculate covariance matrix.
    for i, t in enumerate(t_vector):
        for j, s in enumerate(t_vector):
            if j < i:
                continue
            Rx[i, j] = r_x(t, s)
            Rx[j, i] = Rx[i, j]
    # Apply matrix square root to Rx to find H from (5) in [1].
    # Since Rx is symmetric, its square root will also be symmetric.
    H = sqrtm(Rx)

    # Use only the real part of the matrix if the imaginary side is small.
    # The covariance matrix is symmetric and positive semi-definite and
    # as such it will have a principal square root, i.e. a strictly real
    # square root matrix.
    if (np.linalg.norm(np.imag(H)) / np.linalg.norm(H)) < 1e-6:
        H = np.real(H)

    # Generate Gaussian zero mean stochastic process.
    np.random.seed(random_seed)
    x = np.random.normal(size=(N,))

    # Create process realisation and return all relevant data.
    return H.dot(x), H, Rx


def _q(tau):
    return np.exp(-(tau ** 2) / 2)


def _r(tau, c):
    return np.exp(-((c / 4) * (tau ** 2)) / 2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'lspopt'
DESCRIPTION = "A Python implementation of a multitaper window method for " \
              "estimating Wigner spectra for certain locally stationary processes"
URL = 'https://github.com/hbldh/lspopt'
EMAIL = 'henrik.blidh@gmail.com'
AUTHOR = 'Henrik Blidh'
REQUIRES_PYTHON = '>=2.7.10'
VERSION = None

REQUIRED = [
    'numpy',
    'scipy',
    'six'
]

here = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

with io.open(os.path.join(here, 'CHANGELOG.md'), encoding='utf-8') as f:
    long_description = long_description + '\n' + f.read()

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, NAME, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')
        
        sys.exit()


setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    tests_require=['pytest', ],
    url=URL,
    keywords=['Mathematical Statistics', 'Multitaper', 'Spectrogram'],
    packages=find_packages(exclude=('tests',)),
    package_data={
        'lspopt.data': ['*.npy']
    },
    install_requires=REQUIRED,
    include_package_data=True,
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)
